@@ -15,7 +15,6 @@ kDate_PreviousYear = 1
 kDate_PreviousMonth = 2
 kdate_LastImport =3
 
-
 kCheckedStatus_All =0
 kCheckedStatus_Checked = 1
 kCheckedStatus_UnChecked = 2
@@ -24,6 +23,10 @@ kInOutStatus_All = 0
 kInOutStatus_In = 1
 kInOutStatus_Out = 2
 
+kTagCombo_All = 0
+kTagCombo_With = 1
+kTagCombo_Without = 2
+
 kAccountTypeColumn_AccountTypeId = 0
 kAccountTypeColumn_AccountName = 1
 kAccountTypeColumn_DateField = 2
@@ -35,7 +38,9 @@ kAccountTypeColumn_DateFormat = 7
 
 kTagsColumn_TagId = 0
 kTagsColumn_TagName = 1
-kTagsColumn_UserId = 2
+kTagsColumn_RecordIds = 2
+kTagsColumn_Amount_in = 3
+kTagsColumn_Amount_out = 4
 
 kRecordTagsColumn_RecordTagId = 0
 kRecordTagsColumn_RecordId = 1
